
/*MA****************************************************************/
/*                                                                 */
/*     File: tree.C                                                */
/*   Author: Helmut Schmid                                         */
/*  Purpose:                                                       */
/*  Created: Tue Jan 21 13:27:26 2003                              */
/* Modified: Fri Nov 13 08:23:33 2009 (schmid)                     */
/*                                                                 */
/*ME****************************************************************/

#include <iostream>
using std::cerr;


#include "tree.h"

bool Verbose=true;
bool PrettyPrint=false;
bool UndoMarkov=false;
bool ConvertTraces=false;

FILE *outfile=NULL;

int Trace_Number;


bool is_digit( char c ) { return c >= '0' && c <= '9'; }


/*FA****************************************************************/
/*                                                                 */
/*  index_start                                                    */
/*                                                                 */
/*FE****************************************************************/

bool index_start( const char *s )

{
  if (*s != '-' || !is_digit(*(s+1)))
    return false;
  const char *p;
  for( p=s+2; is_digit(*p); p++) ;
  if (*p == 0 || *p == '-')
    return true;
  else
    return false;
}


/*FA****************************************************************/
/*                                                                 */
/*  last_char                                                      */
/*                                                                 */
/*FE****************************************************************/

char *last_char( char *s ) {
  if (*s) {
    do { s++; } while (*s);
    s--;
  }
  return s;
}

const char *last_char( const char *s ) {
  if (*s) {
    do { s++; } while (*s);
    s--;
  }
  return s;
}



/*FA****************************************************************/
/*                                                                 */
/*  slash_feature_boundary                                         */
/*                                                                 */
/*FE****************************************************************/

static bool slash_feature_boundary( const char *p )

{
  return (*p == '_' || *p == '+' || *p == '=' ||
	  (*p == '%' && *(p-1) != '/') || index_start(p));
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::aux_label                                                */
/*                                                                 */
/*FE****************************************************************/

static bool aux_label( char *s )

{
  if (*(s++) != '<')
    return false;

  while (*s != '[') {
    if (!*s)
      return false;
    s++;
  }

  while (*s != ']') {
    if (!*s)
      return false;
    s++;
  }

  while (*s != '>') {
    if (!*s)
      return false;
    s++;
  }

  return *(s+1) == 0;
}



/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_markup                                               */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_markup( const char *f, char d )

{
  char buffer[1000];
  char *l=label, *r=buffer;
  for(;;) {
    if (*f && (*l == 0 || slash_feature_boundary(l)))
      {
	*(r++) = d;
	while (*f)
	  *(r++) = *(f++) ;
	while ((*(r++) = *(l++))) ;
	break;
      }
    if ((*(r++) = *(l++)) == 0)
      break;
  }
  free(label);
  label = strdup(buffer);
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::undo_markov                                              */
/*                                                                 */
/*FE****************************************************************/

void Tree::undo_markov()

{
  if (daughters.size() > 0) {
    Trees d;
    
    for( size_t i=0; i<daughters.size(); i++ ) {
      Tree *daughter=daughters[i];
      daughter->undo_markov();
      if (aux_label(daughter->label)) {
	for( size_t k=0; k<daughter->daughters.size(); k++ )
	  d.push_back(daughter->daughters[k]);
	daughter->daughters.clear();
	delete daughter;
      }
      else
	d.push_back(daughter);
    }
    daughters = d;
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::slash_cat                                                */
/*                                                                 */
/*  Task: Returns a pointer to the last slash feature (if present) */
/*                                                                 */
/*FE****************************************************************/

char *Tree::slash_cat( char *d )

{
  char *p=last_char(label);
  if (*p == '') {
    int depth=1;
    for( p--; p>label+2; p-- )
      if (*p == '')
	depth++;
      else if (*p == '')
	if (--depth == 0 && *(p-1) == '#' &&
	    (*(p-2) == '+' || *(p-2) == '%'))
	  {
	    *d = *(p-2);
	    return p;
	  }
  }

  return NULL;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::find_trace                                               */
/*                                                                 */
/*FE****************************************************************/

char *Tree::find_trace( char *s, char d, bool &terminal )

{
  char *le=last_char(label);
  char *se=last_char(s);

  // find matching slash feature
  for( ; le > label && *le == *se; le--, se--)
    if (se == s) {
      // matching slash feature recognized
      if (*(--le) == '#') {
	terminal = true;
	le--;
      }
      else
	terminal = false;
      if (*le == d)
	return le;
    }
  
  return NULL;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_trace_feature                                        */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_trace_feature( char c )

{
  char buffer[100];
  sprintf(buffer,"%d",Trace_Number);
  add_markup(buffer,c);
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::delete_slashes_down                                      */
/*                                                                 */
/*FE****************************************************************/

bool Tree::delete_slashes_down( char *sc, char d )

{
  bool term;
  char *trace_start = find_trace( sc, d, term );

  if (trace_start == NULL)
    return false;

  if (term) {
    *trace_start = 0;
    if (d == '_')
      add_trace_feature('-');
    else
      add_trace_feature('=');
    return true;
  }

  for( size_t i=0; i<daughters.size(); i++ )
    if (daughters[i]->delete_slashes_down( sc, d )) {
      *trace_start = 0;
      return true;
    }

  return false;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::delete_slashes_up                                        */
/*                                                                 */
/*FE****************************************************************/

bool Tree::delete_slashes_up( char *sc, char d )

{
  bool term;
  char *trace_start = find_trace( sc, d, term );

  if (trace_start && mother->delete_slashes_up(sc,d)) {
    *trace_start = 0; 
    return true;
  }

  if (d == '_' || d == '+')
    d = '_';
  else
    d = '=';
  for( size_t i=0; i<daughters.size(); i++ )
    if (daughters[i]->delete_slashes_down( sc, d ))
      return true;
  
  return false;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::convert                                                  */
/*                                                                 */
/*FE****************************************************************/

bool Tree::convert( bool &remaining_traces )

{
  bool trace_found=false;
  char d;
  char *p;

  // Check for +#.*
  while ((p=slash_cat(&d)) != NULL) {
    Trace_Number++;
    char *sf = strdup(p);
    bool successful = delete_slashes_up(sf,d);
    free(sf);
    if (!successful) {
      remaining_traces = true;
      Trace_Number--;
      break;
    }
    if (*(p-2)) // slash feature not deleted yet
      *(p-2) = 0;
    trace_found = true;
    add_trace_feature('-');
  }

  for( size_t i=0; i<daughters.size(); i++ )
    trace_found = daughters[i]->convert(remaining_traces) || trace_found;

  return trace_found;
}


/*FA****************************************************************/
/*                                                                 */
/*  quote                                                          */
/*                                                                 */
/*FE****************************************************************/

static char *quote( const char *string )

{
  int i,k;
  static char buffer[10000];
  bool is_trace=false;

  if (string[0] == '(' && *last_char(string) == ')')
    is_trace = true;
  
  for( i=k=0; k<10000; i++, k++) {
    if (!string[i])
      break;
    if (string[i] == '"' || string[i] == '\\'|| string[i] == ' ' || 
	string[i] == '$' || string[i] == '^' || string[i] == '\''||
	string[i] == '[' || string[i] == ']' || string[i] == '{' ||
	string[i] == '}' || string[i] == '=' || string[i] == '<' || 
	string[i] == '>' || string[i] == '#' || string[i] == '/' || 
	(!is_trace && (string[i] == '(' || string[i] == ')')))
      buffer[k++] = '\\';
    buffer[k] = string[i];
  }
  buffer[k] = '\0';

  return &buffer[0];
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::print_subtree                                            */
/*                                                                 */
/*FE****************************************************************/

void Tree::print_subtree( int depth )

{
  if (daughters.size() == 0)
    fprintf(outfile," %s", quote(label));
  else {
    if (PrettyPrint) {
      fputc('\n', outfile);
      for( int i=0; i<depth; i++ )
	fputs("  ", outfile);
    }
    fprintf(outfile, "(%s", quote(label));
    for( size_t i=0; i<daughters.size(); i++ )
      daughters[i]->print_subtree(depth+1);
    fputc(')', outfile);
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::print                                                    */
/*                                                                 */
/*FE****************************************************************/

void Tree::print()

{
  print_subtree(0);
  fputc('\n', outfile);
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::process                                                  */
/*                                                                 */
/*FE****************************************************************/

void Tree::process()

{
  static unsigned long N=0;

  if (Verbose)
    fprintf(stderr,"\r%lu", ++N);

  undo_markov();

  Trace_Number = 0;
  bool remaining_traces;
  do {
    remaining_traces = false;
  } while (convert(remaining_traces) && remaining_traces);

  print();
}
